/*
 *   8OD - Arduino form factor i8086 based SBC
 *   Matthew Millman (tech.mattmillman.com)
 *
 *   Serial flash programming interface
 *
 *   This is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This software is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this software.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdint.h>
#include <stdio.h>
#include <stddef.h>
#include <string.h>
#include <i86.h>

#include "eod_io.h"
#include "uart.h"
#include "flash.h"
#include "pgm.h"
#include "util.h"

#define NUM_NEG_WAITS               20000

#define CMD_RESET                   0x01
#define CMD_BOOT                    0x02
#define CMD_READ_PARAMS             0x03
#define CMD_READ					0x04
#define CMD_ERASE					0x05
#define CMD_WRITE_PAGE				0x06
#define CMD_LOCK_BOOT				0x07
#define CMD_UNLOCK_BOOT				0x08
#define CMD_LOAD_AND_READ           0x09

static int negotiate(void);
static int await_token(uint8_t token, uint8_t reply, uint16_t waitfor);
static int cmd_loop(void);
static void send_data(uint32_t len, void *data);
static void receive_data(uint16_t len, void *data);
static void do_read_params(void);
static void do_read(void);
static void do_write(void);
static void do_erase(void);
static void do_boot_lock(int lock);
static void do_load_and_read(void);
static uint16_t read16(void);
static uint32_t read32(void);

extern int spi_load_and_boot(int);

typedef struct
{
	uint32_t protocol_version;
    uint32_t flash_total_size;
    uint32_t flash_erase_size;
	uint32_t flash_page_size;
    uint32_t flash_boot_offset;
	uint32_t flags;
} sys_info_t;

int pgm_main(void)
{
    int ret;

    ret = cmd_loop();

    uart_close(UARTD);

    return ret;
}

static int cmd_loop(void)
{
    while (1)
    {
        char next_cmd;
        next_cmd = uart_blocking_getc(UARTD);
        switch (next_cmd)
        {
        case CMD_READ_PARAMS:
            do_read_params();
            break;
        case CMD_RESET:
            uart_putc(UARTD, CMD_RESET);
            uart_putc(UARTD, 0x01);
            /* Make sure it's all sent before everything gets put into reset */
            uart_wait_tx(UARTD);
            return PGM_RESET;
            break;
        case CMD_BOOT:
            uart_putc(UARTD, CMD_BOOT);
            uart_putc(UARTD, 0x01);
            return PGM_BOOT;
            break;
		case CMD_READ:
			do_read();
			break;
		case CMD_ERASE:
			do_erase();
			break;
		case CMD_WRITE_PAGE:
			do_write();
			break;
		case CMD_LOCK_BOOT:
			do_boot_lock(1);
			break;
		case CMD_UNLOCK_BOOT:
			do_boot_lock(0);
			break;
        case CMD_LOAD_AND_READ:
            do_load_and_read();
            break;
        }
    }

    return PGM_BOOT;
}

static void do_read_params(void)
{
    sys_info_t info;
    uint8_t *ptr = (uint8_t *)&info;

	info.protocol_version = 1;
    info.flash_total_size = M25P80_SIZE;
    info.flash_erase_size = M25P80_ERASE_SIZE;
	info.flash_page_size = M25P80_PAGE_SIZE;
    info.flash_boot_offset = FLASH_BOOT_OFFSET;

	info.flags = 0;
	if (flash_get_bootarea_lock_state())
		info.flags |= 0x01;

    send_data(sizeof(sys_info_t), ptr);

    uart_putc(UARTD, CMD_READ_PARAMS);
    uart_putc(UARTD, 0x01);
}

static void send_data(uint32_t len, void *data)
{
    uint8_t *ptr = (uint8_t *)data;
   
    while (len--)
    {
        uart_putc(UARTD, *ptr);
        ptr++;
    }
}

static uint16_t read16(void)
{
	uint16_t ret;

    ret = (uint16_t)uart_blocking_getc(UARTD);
    ret |= (uint16_t)uart_blocking_getc(UARTD) << 8;

	return ret;
}

static uint32_t read32(void)
{
	uint32_t ret;

    ret = (uint32_t)uart_blocking_getc(UARTD);
    ret |= (uint32_t)uart_blocking_getc(UARTD) << 8;
    ret |= (uint32_t)uart_blocking_getc(UARTD) << 16;
    ret |= (uint32_t)uart_blocking_getc(UARTD) << 24;

	return ret;
}

static void do_read(void)
{
	uint32_t offset = read32();
	uint32_t len = read32();

	flash_read_to_uart(offset, len, UARTD);

    uart_putc(UARTD, CMD_READ);
	uart_putc(UARTD, 0x01);
}

static void do_write(void)
{
	uint32_t offset = read32();
	uint16_t writeLen = read16();
	/* The writer tool will not send more than this */
	uint8_t page[M25P80_PAGE_SIZE];

	uart_read(UARTD, writeLen, page);

	flash_write(offset, writeLen, page);
	flash_wait_write();

    uart_putc(UARTD, CMD_WRITE_PAGE);
	uart_putc(UARTD, 0x01);
}

static void do_erase(void)
{
	uint32_t offset = read32();
	uint32_t len = read32();

	flash_erase(offset, len);
	flash_wait_write();

    uart_putc(UARTD, CMD_ERASE);
	uart_putc(UARTD, 0x01);
}

static void do_boot_lock(int lock)
{
	if (lock)
		flash_lock_bootarea(1);
	else
		flash_lock_bootarea(0);

	flash_wait_write();

    uart_putc(UARTD, lock ? CMD_LOCK_BOOT : CMD_UNLOCK_BOOT);
	uart_putc(UARTD, 0x01);
}

/* Test function */
static void do_load_and_read(void)
{
    uint16_t i = 0;
    uint8_t far *fp = MK_FP(0x00, 0x0);

    for (;;)
    {
        fp[i] = 0xFF;
        if (i == 0xFFFF)
            break;
        i++;
    }

    i = 0;
    fp = MK_FP(0x1000, 0x0);

    for (;;)
    {
        fp[i] = 0xFF;
        if (i == 0xFFFF)
            break;
        i++;
    }

    if (spi_load_and_boot(0))
    {
        i = 0;
        fp = MK_FP(0x00, 0x0);

        for (;;)
        {
            uart_putc(UARTD, fp[i]);
            if (i == 0xFFFF)
                break;
            i++;
        }

        i = 0;
        fp = MK_FP(0x1000, 0x0);

        for (;;)
        {
            uart_putc(UARTD, fp[i]);
            if (i == 0xFFFF)
                break;
            i++;
        }

        uart_putc(UARTD, CMD_LOAD_AND_READ);
	    uart_putc(UARTD, 0x01);
        return;
    }

    uart_putc(UARTD, CMD_LOAD_AND_READ);
	uart_putc(UARTD, 0x00);
}

/* Establishes connectivity with programmer software */
int pgm_negotiate(void)
{
    int ret;
    char c;

    uart_pgm_open();

    delay_ncycles(100);

    uart_putc(UARTD, '8');
    uart_putc(UARTD, 'O');
    uart_putc(UARTD, 'D');

    delay_ncycles(NUM_NEG_WAITS);
    
    if (uart_getc(UARTD, &c))
    {
        if (c == 0x01)
        {
            // Success
            uart_putc(UARTD, 0x01);
            uart_wait_tx(UARTD);
            ret = 1;
            goto done;
        }
    }

    ret = 0;
    uart_close(UARTD);

done:
    return ret;
}
