﻿using System;
using System.Collections.Generic;
using System.IO;
using System.IO.Ports;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;
using System.Windows.Forms.VisualStyles;

namespace EodPgm
{
    public class EodCommunicationException : Exception
    {
        public EodCommunicationException(string message) : base(message)
        {
            
        }
    }

    public class TargetParameters
    {
        public int ProtocolVersion;
        public int FlashTotalSize;
        public int FlashEraseSize;
        public int FlashPageSize;
        public int FlashBootOffset;
        public bool BootBlockLocked;
    }

    public enum OperationType
    {
        Write,
        Erase,
        Read
    }

    public class EodIsp
    {
        public delegate void ConnectedCallback(bool success, string failMessage);
        public event ConnectedCallback Connected;

        public delegate void OperationProgressCallback(int percent);
        public event OperationProgressCallback OperationProgress;

        public delegate void OperationChangedCallback(OperationType type);
        public event OperationChangedCallback OperationChanged;

        public delegate void OperationCompletedCallback(bool success, string failMessage);
        public event OperationCompletedCallback OperationResult;


        public TargetParameters TargetParameters{ get { return _targetParameters; } }
        public byte[] ReadData { get { return _data; } }

        private SerialPort _port;
        private TargetParameters _targetParameters;
        private byte[] _data;
        private int _len;
        private int _offset;

        private const byte CMD_RESET = 0x01;
        private const byte CMD_BOOT = 0x02;
        private const byte CMD_READ_PARAMS = 0x03;
        private const byte CMD_READ = 0x04;
        private const byte CMD_ERASE = 0x05;
        private const byte CMD_WRITE_PAGE = 0x06;
        private const byte CMD_LOCK_BOOT = 0x07;
        private const byte CMD_UNLOCK_BOOT = 0x08;
        private const byte CMD_LOAD_AND_READ = 0x09;

        public EodIsp(SerialPort port)
        {
            _port = port;
        }
        
        public void Write(byte[] data, int offset, int len)
        {
            if ((offset + len) > _targetParameters.FlashTotalSize)
            {
                if (OperationResult != null)
                    OperationResult(false, "Requested write region is beyond extend of target device");
                return;
            }

            if ((_offset + len > _targetParameters.FlashBootOffset) && _targetParameters.BootBlockLocked)
            {
                if (OperationResult != null)
                    OperationResult(false, "Operation would overwrite boot block, but it is locked");
                return;
            }

            var writeThread = new Thread(WriteThread);
            
            _len = len;
            _offset = offset;
            _data = data;

            writeThread.Start();
        }

        public void Read(int offset, int len)
        {
            if ((offset + len) > _targetParameters.FlashTotalSize)
            {
                if (OperationResult != null)
                    OperationResult(false, "Requested read region is beyond extend of target device");
                return;
            }

            var readThread = new Thread(ReadThread);
            _len = len;
            _offset = offset;
            readThread.Start();
        }

        public void Erase(int offset, int len)
        {
            if ((offset + len) > _targetParameters.FlashTotalSize)
            {
                if (OperationResult != null)
                    OperationResult(false, "Requested erase region is beyond extend of target device");
                return;
            }

            if ((_offset + len > _targetParameters.FlashBootOffset) && _targetParameters.BootBlockLocked)
            {
                if (OperationResult != null)
                    OperationResult(false, "Operation would erase boot block, but it is locked");
                return;
            }

            var eraseThread = new Thread(EraseThread);
            _len = len;
            _offset = offset;
            eraseThread.Start();
        }

        public void LoadAndRead()
        {
            var writeThread = new Thread(LoadAndReadThread);
            _len = 0x20000;
            writeThread.Start();
        }


        public bool SetBootBlockLock(bool locked)
        {
            IEnumerable<byte> toSend = new byte[] { locked ? CMD_LOCK_BOOT : CMD_UNLOCK_BOOT };

            _port.Write(toSend.ToArray(), 0, toSend.Count());

            /* Wait for the completion byte */
            bool ret = CheckReturnCode(locked ? CMD_LOCK_BOOT : CMD_UNLOCK_BOOT);

            ReadTargetParameters();

            return ret;
        }

        /* Hard reset */
        public bool ResetTarget()
        {
            IEnumerable<byte> toSend = new byte[] { CMD_RESET };
            _port.Write(toSend.ToArray(), 0, toSend.Count());

            return CheckReturnCode(CMD_RESET);
        }

        /* Exit loader and boot */
        public bool BootTarget()
        {
            IEnumerable<byte> toSend = new byte[] { CMD_BOOT };
            _port.Write(toSend.ToArray(), 0, toSend.Count());

            return CheckReturnCode(CMD_BOOT);
        }

        private void WriteThread()
        {
            if (OperationChanged != null)
                OperationChanged(OperationType.Erase);

            if (!DoErase())
            {
                if (OperationResult != null)
                    OperationResult(false, "Failed to erase requested region");
                return;
            }

            if (OperationChanged != null)
                OperationChanged(OperationType.Write);


            int written = 0;
            UInt32 offset = (UInt32)_offset;

            while (written < _len)
            {
                UInt16 thisPageLen = (UInt16)((_len - written) > _targetParameters.FlashPageSize
                    ? (int)_targetParameters.FlashPageSize
                    : (_len - written));

                IEnumerable<byte> toSend = new byte[] { CMD_WRITE_PAGE };

                toSend = toSend.Concat(BitConverter.GetBytes(offset));
                toSend = toSend.Concat(BitConverter.GetBytes(thisPageLen));
                toSend = toSend.Concat(_data.Skip(written).Take(thisPageLen));

                _port.Write(toSend.ToArray(), 0, toSend.Count());

                written += thisPageLen;
                offset += thisPageLen;

                if (!CheckReturnCode(CMD_WRITE_PAGE))
                {
                    if (OperationResult != null)
                        OperationResult(false, "Invalid response received from target while writing page");
                    return;
                }

                if (OperationProgress != null)
                    OperationProgress((written * 100) / _len);
            }

            OperationResult(true, null);
        }

        private void ReadThread()
        {
            if (OperationChanged != null)
                OperationChanged(OperationType.Read);

            IEnumerable<byte> toSend = new byte[] { CMD_READ };

            toSend = toSend.Concat(BitConverter.GetBytes((UInt32)_offset));
            toSend = toSend.Concat(BitConverter.GetBytes((UInt32)_len));

            _port.Write(toSend.ToArray(), 0, toSend.Count());

            _data = new byte[_len];

            int readSize = (int)_targetParameters.FlashPageSize;
            int read = 0;
            while (read < _len)
            {
                int thisRead = (_len - read) > readSize ? readSize : (_len - read);
                Buffer.BlockCopy(SynchronousRead(thisRead), 0, _data, read, thisRead);
                read += thisRead;
                if (OperationProgress != null)
                    OperationProgress((read * 100) / _len);
            }

            if (CheckReturnCode(CMD_READ))
            {
                if (OperationResult != null)
                    OperationResult(true, null);
            }
            else
            {
                OperationResult(false, "Read failed");
            }
        }

        private void LoadAndReadThread()
        {
            if (OperationChanged != null)
                OperationChanged(OperationType.Read);

            IEnumerable<byte> toSend = new byte[] { CMD_LOAD_AND_READ };

            _port.Write(toSend.ToArray(), 0, toSend.Count());

            _data = new byte[_len];

            int readSize = 1024;
            int read = 0;
            while (read < _len)
            {
                int thisRead = (_len - read) > readSize ? readSize : (_len - read);
                Buffer.BlockCopy(SynchronousRead(thisRead), 0, _data, read, thisRead);
                read += thisRead;
                if (OperationProgress != null)
                    OperationProgress((read * 100) / _len);
            }

            if (CheckReturnCode(CMD_LOAD_AND_READ))
            {
                if (OperationResult != null)
                    OperationResult(true, null);
            }
            else
            {
                OperationResult(false, "Read failed");
            }
        }


        private void EraseThread()
        {
            if (OperationChanged != null)
                OperationChanged(OperationType.Erase);

            if (!DoErase())
            {
                if (OperationResult != null)
                    OperationResult(false, "Erase failed");
            }

            if (OperationResult != null)
                OperationResult(true, null);
        }

        private bool DoErase()
        {
            int offset = _offset;
            int erased = 0;

            while (erased < _len)
            {
                int blkStart = offset & ~(_targetParameters.FlashEraseSize - 1); 

                IEnumerable<byte> toSend = new byte[] {CMD_ERASE};

                toSend = toSend.Concat(BitConverter.GetBytes((UInt32)blkStart));
                toSend = toSend.Concat(BitConverter.GetBytes((UInt32)_targetParameters.FlashEraseSize));

                _port.Write(toSend.ToArray(), 0, toSend.Count());

                if (!CheckReturnCode(CMD_ERASE))
                    return false;

                if (erased + _targetParameters.FlashEraseSize > _len)
                {
                    erased = _len;
                    offset = (_offset + _len);
                }
                else
                {
                    erased += _targetParameters.FlashEraseSize;
                    offset += _targetParameters.FlashEraseSize;
                }

                if (OperationProgress != null)
                    OperationProgress((erased * 100) / _len);
            }

            return true;
        }

        private bool ReadTargetParameters()
        {
            _port.Write(new byte[] { CMD_READ_PARAMS }, 0, 1);

            byte[] raw = SynchronousRead(6 * sizeof(UInt32));

            _targetParameters = new TargetParameters();
            _targetParameters.ProtocolVersion = (int)BitConverter.ToUInt32(raw, sizeof(UInt32) * 0);
            _targetParameters.FlashTotalSize = (int)BitConverter.ToUInt32(raw, sizeof(UInt32) * 1);
            _targetParameters.FlashEraseSize = (int)BitConverter.ToUInt32(raw, sizeof(UInt32) * 2);
            _targetParameters.FlashPageSize = (int)BitConverter.ToUInt32(raw, sizeof(UInt32) * 3);
            _targetParameters.FlashBootOffset = (int)BitConverter.ToUInt32(raw, sizeof(UInt32) * 4);
            _targetParameters.BootBlockLocked = (int)BitConverter.ToUInt32(raw, sizeof(UInt32) * 5) == 0x01;

            return CheckReturnCode(CMD_READ_PARAMS);
        }

        private bool CheckReturnCode(byte cmd)
        {
            byte executedCmd = (byte)_port.ReadByte();
            if (executedCmd != cmd)
                return false;

            byte result = (byte)_port.ReadByte();

            return (result == 0x01);
        }

        private byte[] SynchronousRead(int len)
        {
            byte[] buffer = new byte[len];
            int total = 0;
            while (len > 0)
            {
                try
                {
                    int read = _port.Read(buffer, total, len);
                    len -= read;
                    total += read;

                    if (len > 0)
                        Thread.Sleep(1);
                }
                catch (IOException)
                {
                    return null;
                }
                catch (InvalidOperationException)
                {
                    return null;
                }
            }

            return buffer;
        }

        public void Connect()
        {
            var connectThread = new Thread(ConnectThread);
            connectThread.Start();
        }

        private void ConnectThread()
        {
            try
            {
                _port.Open();

                if (!TargetHandshake())
                    return;

                Thread.Sleep(100);

                ReadTargetParameters();

                Connected(true, null);
            }
            catch (UnauthorizedAccessException ex)
            {
                if (Connected != null)
                    Connected(false, ex.Message);
            }
        }

        private bool TargetHandshake()
        {
            int handshakeState = 0;

            for (;;)
            {
                try
                {
                    var toRead = _port.BytesToRead;
                    if (toRead >= 1)
                    {
                        byte read = (byte) _port.ReadByte();
                        switch (handshakeState)
                        {
                            case 0:
                                if (read == '8')
                                    handshakeState = 1;
                                else
                                    handshakeState = 0;
                                break;
                            case 1:
                                if (read == 'O')
                                    handshakeState = 2;
                                else
                                    handshakeState = 0;
                                break;
                            case 2:
                                if (read == 'D')
                                {
                                    handshakeState = 3;
                                    _port.Write(new byte[] { 0x01 }, 0, 1);
                                }
                                else
                                {
                                    handshakeState = 0;
                                }
                                break;
                            case 3:
                                if (read == 0x01)
                                    handshakeState = 4;
                                else
                                    handshakeState = 0;
                                break;

                            default:
                                handshakeState = 0;
                                break;
                        }

                        if (handshakeState == 4)
                            break;
                    }
                }
                catch (InvalidOperationException ex)
                {
                    Connected(false, null);
                    return false;
                }
            }

            return true;
        }
    }
}
