﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Globalization;
using System.IO;
using System.IO.Ports;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using ConsoleApplication1;

namespace EodPgm
{
    public enum FormState
    {
        Disconnected,
        Connecting,
        Connected,
        OperationInProgress
    }

    public partial class EodPgmForm : Form
    {
        private string _serialPort;
        private FormState _state;
        private OperationType _lastOperationType;
        private EodIsp _isp;
        private byte[] _hexToWrite;
        private byte[] _binToWrite;
        private byte[] _expandedExecutable;
        private byte[] _binJustRead;

        public EodPgmForm()
        {
            InitializeComponent();

            _state = FormState.Disconnected;
        }

        private void EodPgmForm_Load(object sender, EventArgs e)
        {
            serialPort.BaudRate = 115200;
            serialPort.Handshake = Handshake.None;
            serialPort.DataBits = 8;
            serialPort.StopBits = StopBits.One;

            var namesList = SerialPort.GetPortNames().ToList();
            namesList.Sort();

            ddlSerialPorts.Items.AddRange(namesList.ToArray());

            SelectComPort();

            txtWriteOffset.Text = string.Format("{0:X}", Properties.Settings.Default.LastWriteOffset);
            txtReadOffset.Text = string.Format("{0:X}", Properties.Settings.Default.LastReadOffset);
            txtReadLen.Text = string.Format("{0:X}", Properties.Settings.Default.LastReadLen);

            UpdateForm();
        }

        private void SaveDefaults()
        {
            int writeOffset;
            int readOffset;
            int readLen;

            if (int.TryParse(txtWriteOffset.Text, NumberStyles.HexNumber, null, out writeOffset))
            {
                Properties.Settings.Default.LastWriteOffset = writeOffset;
            }

            if (int.TryParse(txtReadOffset.Text, NumberStyles.HexNumber, null, out readOffset))
            {
                Properties.Settings.Default.LastReadOffset = readOffset;
            }

            if (int.TryParse(txtReadLen.Text, NumberStyles.HexNumber, null, out readLen))
            {
                Properties.Settings.Default.LastReadLen = readLen;
            }

            Properties.Settings.Default.Save();
        }

        private void SelectComPort()
        {
            if (string.IsNullOrEmpty(Properties.Settings.Default.ComPort))
            {
                if (ddlSerialPorts.Items.Count > 0)
                    ddlSerialPorts.SelectedIndex = 0;

                Properties.Settings.Default.ComPort = ddlSerialPorts.Text;
                Properties.Settings.Default.Save();
            }
            else
            {
                if (ddlSerialPorts.Items.Contains(Properties.Settings.Default.ComPort))
                {
                    ddlSerialPorts.SelectedIndex = ddlSerialPorts.Items.IndexOf(Properties.Settings.Default.ComPort);
                }
                else
                {
                    if (ddlSerialPorts.Items.Count > 0)
                        ddlSerialPorts.SelectedIndex = 0;
                }
            }
        }

        private void ddlSerialPorts_SelectedIndexChanged(object sender, EventArgs e)
        {
            Properties.Settings.Default.ComPort = ddlSerialPorts.Text;
            Properties.Settings.Default.Save();

            _serialPort = ddlSerialPorts.Text;
        }

        private void btnEnterPgm_Click(object sender, EventArgs e)
        {
            if (string.IsNullOrEmpty(_serialPort))
            {
                MessageBox.Show("No serial port selected", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            serialPort.PortName = _serialPort;
            _isp = new EodIsp(serialPort);
            _isp.Connected += IspOnConnected;
            _isp.OperationResult += IspOnOperationResult;
            _isp.OperationProgress += IspOnOperationProgress;
            _isp.OperationChanged += IspOnOperationChanged;
            _isp.Connect();
            _state = FormState.Connecting;
            UpdateForm();
        }

        private void IspOnOperationChanged(OperationType type)
        {
            if (InvokeRequired)
            {
                BeginInvoke(new EodIsp.OperationChangedCallback(IspOnOperationChanged),
                    new object[] { type });
                return;
            }

            _lastOperationType = type;
            UpdateForm();
        }

        private void IspOnOperationProgress(int percent)
        {
            if (InvokeRequired)
            {
                BeginInvoke(new EodIsp.OperationProgressCallback(IspOnOperationProgress),
                    new object[] { percent });
                return;
            }

            prgOperation.Value = percent;
        }

        private void IspOnOperationResult(bool success, string failMessage)
        {
            if (InvokeRequired)
            {
                BeginInvoke(new EodIsp.OperationCompletedCallback(IspOnOperationResult),
                    new object[] {success, failMessage});
                return;
            }

            if (!success)
            {
                MessageBox.Show(failMessage, "Operation failed", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            else
            {
                if (_lastOperationType == OperationType.Read)
                    _binJustRead = _isp.ReadData;
            }

            _state = FormState.Connected;

            UpdateForm();
        }

        private void IspOnConnected(bool success, string failMessage)
        {
            if (InvokeRequired)
            {
                BeginInvoke(new EodIsp.ConnectedCallback(IspOnConnected), new object [] { success, failMessage });
                return;
            }

            if (success)
            {
                _state = FormState.Connected;
                lblFlashSize.Text = string.Format("Flash size: {0}", BytesToString(_isp.TargetParameters.FlashTotalSize));
                lblBootOffset.Text = string.Format("Boot area offset: 0x{0:X5}", _isp.TargetParameters.FlashBootOffset);
            }
            else
            {
                _state = FormState.Disconnected;
                serialPort.Close();
                if (failMessage != null)
                    MessageBox.Show(failMessage, "Failed to connect", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }

            UpdateForm();
        }

        private void UpdateForm()
        {
            if (_state == FormState.Disconnected)
            {
                btnBoot.Enabled = false;
                btnReset.Enabled = false;
                ddlSerialPorts.Enabled = true;
                btnEnterPgm.Enabled = true;
                btnResetApp.Enabled = false;

                grpControl.Enabled = true;
                grpProgramBoot.Enabled = false;
                grpWriteBin.Enabled = false;
                grpReadBin.Enabled = false;

                lblStatus.Text = "Status: Disconnected";
            }
            else if (_state == FormState.Connecting)
            {
                btnBoot.Enabled = false;
                btnReset.Enabled = false;
                ddlSerialPorts.Enabled = false;
                btnEnterPgm.Enabled = false;
                btnResetApp.Enabled = true;

                grpControl.Enabled = false;
                grpProgramBoot.Enabled = false;
                grpWriteBin.Enabled = false;
                grpReadBin.Enabled = false;

                lblStatus.Text = "Status: Waiting for manual target reset";
            }
            else if (_state == FormState.Connected)
            {
                btnBoot.Enabled = true;
                btnReset.Enabled = true;
                btnEnterPgm.Enabled = false;
                ddlSerialPorts.Enabled = false;
                btnResetApp.Enabled = true;

                grpControl.Enabled = true;
                grpProgramBoot.Enabled = true;
                grpWriteBin.Enabled = true;
                grpReadBin.Enabled = true;

                if (_binToWrite != null)
                    btnWriteBin.Enabled = true;
                else
                    btnWriteBin.Enabled = false;

                if (_hexToWrite != null)
                    btnProgramHexFile.Enabled = true;
                else
                    btnProgramHexFile.Enabled = false;

                if (_binJustRead != null)
                    btnSaveBin.Enabled = true;
                else
                    btnSaveBin.Enabled = false;

                lblStatus.Text = "Status: Connected";
            }
            else if (_state == FormState.OperationInProgress)
            {
                grpControl.Enabled = false;
                grpProgramBoot.Enabled = false;
                grpWriteBin.Enabled = false;
                grpReadBin.Enabled = false;
                btnResetApp.Enabled = false;

                switch (_lastOperationType)
                {
                    case OperationType.Read:
                        lblStatus.Text = "Status: Read in progress";
                        break;
                    case OperationType.Write:
                        lblStatus.Text = "Status: Write in progress";
                        break;
                    case OperationType.Erase:
                        lblStatus.Text = "Status: Erase in progress";
                        break;
                }
            }
        }

        static string BytesToString(long byteCount)
        {
            string[] suf = { "B", "KB", "MB", "GB", "TB", "PB", "EB" };
            if (byteCount == 0)
                return "0" + suf[0];
            long bytes = Math.Abs(byteCount);
            int place = Convert.ToInt32(Math.Floor(Math.Log(bytes, 1024)));
            double num = Math.Round(bytes / Math.Pow(1024, place), 1);
            return (Math.Sign(byteCount) * num).ToString() + suf[place];
        }

        private void btnOpenHexFile_Click(object sender, EventArgs e)
        {
            OpenFileDialog dlg = new OpenFileDialog();
            dlg.Title = "Open HEX file";
            dlg.Filter = "HEX files|*.hex";
            if (!string.IsNullOrEmpty(Properties.Settings.Default.LastHexOpenDir))
                dlg.InitialDirectory = Properties.Settings.Default.LastHexOpenDir;

            if (dlg.ShowDialog() == DialogResult.OK)
            {
                Properties.Settings.Default.LastHexOpenDir = Path.GetDirectoryName(dlg.FileName);
                Properties.Settings.Default.Save();

                try
                {
                    var ihex = new IntelHex();
                    ihex.Process(dlg.FileName);
                    _hexToWrite = ihex.Image;
                    _expandedExecutable = ihex.MakeBinImage(0x70000, 0xFF);
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.Message, "Error opening HEX file", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return;
                }

                UpdateForm();
            }
        }

        private void btnProgramHexFile_Click(object sender, EventArgs e)
        {
            _isp.Write(_hexToWrite, _isp.TargetParameters.FlashBootOffset, _hexToWrite.Length);
            _state = FormState.OperationInProgress;
            prgOperation.Value = 0;
            UpdateForm();
            SaveDefaults();
        }

        private void btnOpenBin_Click(object sender, EventArgs e)
        {
            OpenFileDialog dlg = new OpenFileDialog();
            dlg.Title = "Open BIN file";
            dlg.Filter = "BIN files|*.bin|All files|*.*";
            if (!string.IsNullOrEmpty(Properties.Settings.Default.LastBinOpenDir))
                dlg.InitialDirectory = Properties.Settings.Default.LastBinOpenDir;

            if (dlg.ShowDialog() == DialogResult.OK)
            {
                Properties.Settings.Default.LastBinOpenDir = Path.GetDirectoryName(dlg.FileName);
                Properties.Settings.Default.Save();

                var info = new FileInfo(dlg.FileName);
                if (info.Length > _isp.TargetParameters.FlashTotalSize)
                {
                    MessageBox.Show("Selected file is larger than target flash device", "Error", MessageBoxButtons.OK,
                        MessageBoxIcon.Error);
                    return;
                }

                _binToWrite = File.ReadAllBytes(dlg.FileName);
                UpdateForm();
            }
        }

        private void btnWriteBin_Click(object sender, EventArgs e)
        {
            int offset;
            if (!int.TryParse(txtWriteOffset.Text, NumberStyles.HexNumber, null, out offset))
            {
                MessageBox.Show("Invalid offset entered", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            _isp.Write(_binToWrite, offset, _binToWrite.Length);
            _state = FormState.OperationInProgress;
            prgOperation.Value = 0;
            UpdateForm();
            SaveDefaults();
        }

        private void btnReadBin_Click(object sender, EventArgs e)
        {
            int offset;
            if (!int.TryParse(txtReadOffset.Text, NumberStyles.HexNumber, null, out offset))
            {
                MessageBox.Show("Invalid offset entered", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            int length;
            if (!int.TryParse(txtReadLen.Text, NumberStyles.HexNumber, null, out length))
            {
                MessageBox.Show("Invalid length entered", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            if ((offset + length) > _isp.TargetParameters.FlashTotalSize)
            {
                MessageBox.Show("Specified read is larger than target flash device", "Error", MessageBoxButtons.OK,
                    MessageBoxIcon.Error);
                return;
            }

            _isp.Read(offset, length);

            _state = FormState.OperationInProgress;
            prgOperation.Value = 0;
            UpdateForm();
            SaveDefaults();
        }

        private void btnSaveBin_Click(object sender, EventArgs e)
        {
            SaveFileDialog dlg = new SaveFileDialog();
            dlg.Title = "Save BIN file";
            dlg.Filter = "BIN files|*.bin|All files|*.*";

            if (!string.IsNullOrEmpty(Properties.Settings.Default.LastSaveName))
                dlg.FileName = Properties.Settings.Default.LastSaveName;
            else
                dlg.FileName = "data.bin";

            if (!string.IsNullOrEmpty(Properties.Settings.Default.LastSaveDir))
                dlg.InitialDirectory = Properties.Settings.Default.LastSaveDir;

            if (dlg.ShowDialog() == DialogResult.OK)
            {
                Properties.Settings.Default.LastSaveDir = Path.GetDirectoryName(dlg.FileName);
                Properties.Settings.Default.LastSaveName = Path.GetFileName(dlg.FileName);
                Properties.Settings.Default.Save();

                try
                {
                    File.WriteAllBytes(dlg.FileName, _binJustRead);
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.Message, "Error saving file", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }
        }

        private void btnBoot_Click(object sender, EventArgs e)
        {
            if (_isp.BootTarget())
            {
                _state = FormState.Disconnected;
                UpdateForm();
                serialPort.Close();
            }
        }

        private void btnReset_Click(object sender, EventArgs e)
        {
            if (_isp.ResetTarget())
            {
                _state = FormState.Disconnected;
                UpdateForm();
                serialPort.Close();
            }
        }

        private void btnResetApp_Click(object sender, EventArgs e)
        {
            _state = FormState.Disconnected;
            serialPort.Close();
            lblFlashSize.Text = "Flash size: (Unknown)";
            lblBootOffset.Text = "Boot area offset: (Unknown)";
            UpdateForm();
        }

        private void btnSaveExecutable_Click(object sender, EventArgs e)
        {
            _isp.LoadAndRead();
            _state = FormState.OperationInProgress;
            prgOperation.Value = 0;
            UpdateForm();
        }
    }
}
