﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Runtime.Serialization;
using System.Text;
using System.Threading.Tasks;

namespace ConsoleApplication1
{
    public enum IntelHexRecordType
    {
        Data = 0x00,
        EOF = 0x01,
        ExtendedSegmentAddress = 0x02,
        StartSegmentAddress = 0x03,
        ExtededLinearAddress = 0x04,
        StartLinearAddress = 0x05
    }

    [Serializable]
    public class IntelHexParseException : Exception
    {
        public IntelHexParseException()
        {
        }

        public IntelHexParseException(string message) : base(message)
        {
        }

        public IntelHexParseException(string message, Exception inner) : base(message, inner)
        {
        }

        protected IntelHexParseException(
            SerializationInfo info,
            StreamingContext context) : base(info, context)
        {
        }
    }

    public class IntelHex
    {
        private class IntelHexLine
        {
            public IntelHexRecordType Type { get; set; }
            public ushort Address { get; set; }
            public byte[] Data { get; set; }
            public ushort ExtendedSegmentAddress { get; set; }
        }

        private class DataSegment
        {
            public ushort SegmentAddress { get; set; }
            public ushort SegmentOffset { get; set; }
            public int PhysicalAddress { get { return (SegmentAddress << 4) + SegmentOffset; }}
            public byte[] Data { get; set; }
        }

        private List<IntelHexLine> _lines;
        private List<DataSegment> _segments;
        private byte[] _executable;

        private const ushort ValidSegmentMagic = 0x6501;
        private const ushort EndOfSegmentsMagic = 0x6502;

        public IntelHex()
        {
            _lines = new List<IntelHexLine>();
            _segments = new List<DataSegment>();
        }

        public byte[] Image { get { return _executable; }}

        public void Process(string fileName)
        {
            ParseFile(fileName);
            Collapse();
            Align();
            CheckSequence();
            MakeTargetExecutable();
        }

        /* Parse iHex file */
        private void ParseFile(string fileName)
        {
            int lineNum = 0;
            var lines = File.ReadAllLines(fileName);
            foreach (var line in lines)
            {
                if (string.IsNullOrEmpty(line.Trim()))
                    continue;

                lineNum++;
                if (line[0] != ':')
                    throw new IntelHexParseException(string.Format("Line: {0}: Line doesn't start with ':'", lineNum));
                string lengthStr = line.Substring(1, 2);
                byte length;
                if (!byte.TryParse(lengthStr, NumberStyles.HexNumber, null, out length))
                    throw new IntelHexParseException(string.Format("Line: {0}: Can't parse record type ({1})", lineNum, lengthStr));
                if (length > 0x10)
                    throw new IntelHexParseException(string.Format("Line: {0}: Over length line", lineNum));
                if (line.Length > (length * 2) + 11)
                    throw new IntelHexParseException(string.Format("Line: {0}: Line is longer than specified", lineNum));
                List<byte> checksumBytes = new List<byte>();
                for (int i = 1; i < (line.Length - 2); i += 2)
                {
                    byte b;
                    if (!byte.TryParse(line.Substring(i, 2), NumberStyles.HexNumber, null, out b))
                        throw new IntelHexParseException(string.Format("Line: {0}: Error parsing line for checksum", lineNum));
                    checksumBytes.Add(b);
                }
                var checkSum = Checksum8(checksumBytes.ToArray(), checksumBytes.Count);
                var statedChecksumStr = line.Substring(line.Length - 2, 2);
                byte statedChecksum;
                if (!byte.TryParse(statedChecksumStr, NumberStyles.HexNumber, null, out statedChecksum))
                    throw new IntelHexParseException(string.Format("Line: {0}: Error parsing checksum byte", lineNum));
                if (checkSum != statedChecksum)
                    throw new IntelHexParseException(string.Format("Line: {0}: Incorrect checksum", lineNum));
                string addrStr = line.Substring(3, 4);
                ushort addr;
                if (!ushort.TryParse(addrStr, NumberStyles.HexNumber, null, out addr))
                    throw new IntelHexParseException(string.Format("Line: {0}: Can't parse record address ({1})", lineNum, addrStr));
                var typeStr = line.Substring(7, 2);
                byte type;
                if (!byte.TryParse(typeStr, NumberStyles.HexNumber, null, out type))
                    throw new IntelHexParseException(string.Format("Line: {0}: Can't parse record type ({1})", lineNum, typeStr));
                if (type > (byte)IntelHexRecordType.StartLinearAddress)
                    throw new IntelHexParseException(string.Format("Line: {0}: Invalid record type ({1})", lineNum, type));
                var iline = new IntelHexLine();
                iline.Type = (IntelHexRecordType)type;
                iline.Address = addr;
                if (iline.Type == IntelHexRecordType.Data)
                {
                    List<byte> bytes = new List<byte>();
                    for (int i = 0; i < length; i ++)
                    {
                        var byteStr = line.Substring(9 + (i * 2), 2);
                        byte thisByte;
                        if (!byte.TryParse(byteStr, NumberStyles.HexNumber, null, out thisByte))
                            throw new IntelHexParseException(string.Format("Line: {0}: Can't parse record data", lineNum));
                        bytes.Add(thisByte);
                    }
                    iline.Data = bytes.ToArray();
                }
                else if (iline.Type == IntelHexRecordType.ExtendedSegmentAddress)
                {
                    string extendedSegAddrStr = line.Substring(9, 4);
                    ushort extendedSeg;
                    if (!ushort.TryParse(extendedSegAddrStr, NumberStyles.HexNumber, null, out extendedSeg))
                        throw new IntelHexParseException(string.Format("Line: {0}: Can't parse extended segment address ({1})", lineNum, extendedSeg));
                    iline.ExtendedSegmentAddress = extendedSeg;
                }
                else if (iline.Type == IntelHexRecordType.EOF) { }
                else
                {
                    throw new IntelHexParseException(string.Format("Line: {0}: Can't parse record type ({1})", lineNum, iline.Type));
                }
                _lines.Add(iline);
            }

            if (!_lines.Any(line => line.Type == IntelHexRecordType.EOF))
                throw new IntelHexParseException("No EOF record observed");
        }

        private byte Checksum8(byte[] buf, int len)
        {
            byte i;
            byte result = 0;
            for (i = 0; i < len; i++)
            {
                unchecked
                {
                    result += buf[i];
                }
            }
            return (byte)((result ^ 0xFF) + 1);
        }

        /* Collapse into unique contiguous segments */
        private void Collapse()
        {
            ushort currentSegAddr = 0;
            int endOfLastChunk = 0;

            DataSegment segment = new DataSegment();

            foreach (var line in _lines)
            {
                if (line.Type == IntelHexRecordType.ExtendedSegmentAddress)
                {
                    _segments.Add(segment);
                    segment = new DataSegment();
                    segment.SegmentAddress = line.ExtendedSegmentAddress;
                    currentSegAddr = line.ExtendedSegmentAddress;
                    endOfLastChunk = 0;
                }

                if (line.Type == IntelHexRecordType.Data)
                {
                    if (segment.Data == null)
                    {
                        segment.SegmentOffset = line.Address;
                        segment.Data = line.Data;
                        endOfLastChunk = line.Address + line.Data.Count();
                    }
                    else
                    {
                        if (line.Address > endOfLastChunk)
                        {
                            _segments.Add(segment);
                            segment = new DataSegment();
                            segment.SegmentAddress = currentSegAddr;
                            segment.SegmentOffset = line.Address;
                            segment.Data = line.Data;
                            endOfLastChunk = line.Address + line.Data.Count();
                        }
                        else
                        {
                            segment.Data = segment.Data.Concat(line.Data).ToArray();
                            endOfLastChunk = line.Address + line.Data.Count();
                        }
                    }
                }
                if (line.Type == IntelHexRecordType.EOF)
                {
                    _segments.Add(segment);
                    segment = null;
                }
            }

            _segments.RemoveAll(s => s.Data == null);
        }

        private void Align()
        {
            /* Pad everything to 16 bits to help the boot process */
            foreach (var segment in _segments)
            {
                if ((segment.SegmentOffset % 2) != 0)
                    throw new IntelHexParseException(string.Format("Unaligned segment found! @ {0:4X}:{1:4X}",
                        segment.SegmentAddress, segment.SegmentOffset));

                if ((segment.Data.Length % 2) != 0)
                {
                    byte[] final = new byte[segment.Data.Length + 1];
                    Buffer.BlockCopy(segment.Data, 0, final, 0, segment.Data.Length);
                    final[segment.Data.Length] = 0x00;
                    segment.Data = final;
                }
            }
        }

        private void CheckSequence()
        {
            int lastSegEnd = 0;
            foreach (var segment in _segments)
            {
                if (segment.PhysicalAddress + segment.Data.Length > 0x70000)
                    throw new Exception("Can't make image because it contains data overlapping bootrom RAM area (0x70000-0x7FFFF)");
                if (segment.PhysicalAddress < lastSegEnd)
                    throw new Exception(string.Format("Overlapping segment found! @ {0:4X}:{1:4X}",
                        segment.SegmentAddress, segment.SegmentOffset));
                if (segment.Data.Length < 2)
                    throw new Exception(string.Format("Segment less than 2 bytes found @ {0:4X}:{1:4X}",
                        segment.SegmentAddress, segment.SegmentOffset));

                lastSegEnd = segment.PhysicalAddress + segment.Data.Length;
            }
        }

        private void MakeTargetExecutable()
        {
            var ms = new MemoryStream();
            var writer = new BinaryWriter(ms);
            ushort numSections = (ushort)_segments.Count;

            foreach (var segment in _segments)
            {
                if ((segment.Data.Length % 2) != 0)
                {
                    throw new Exception("Internal error - segment not aligned!");
                }

                writer.Write(ValidSegmentMagic);
                writer.Write(segment.SegmentAddress);
                writer.Write(segment.SegmentOffset);
                writer.Write((ushort)segment.Data.Count());
                writer.Write(segment.Data);
            }

            writer.Write(EndOfSegmentsMagic);

            var binReader = new BinaryReader(writer.BaseStream);
            binReader.BaseStream.Position = 0;
            _executable = binReader.ReadBytes((int)ms.Length);
        }

        /* Test function - simulates what happens on boot, to test
         * the integrity of this code.
         * 
         * Seen that big chunk of assembler in spiloader.asm ?
         * This function is all it does.
         * 
         * Call with size=0x70000 and fillChar=0x00 and this is
         * what will end up in the target's RAM.
         * 
         */
        public byte[] MakeBinImage(int size, byte fillChar)
        {
            byte[] image = Enumerable.Repeat(fillChar, size).ToArray();

            foreach (var segment in _segments)
                Buffer.BlockCopy(segment.Data, 0, image, segment.PhysicalAddress, segment.Data.Length);

            return image;
        }
    }
}
